<?php

namespace Drupal\dbxschema\Database;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\dbxschema\Exception\ConnectionException;
use Drupal\dbxschema\Exception\DatabaseToolException;

/**
 * Database tool.
 *
 * This class provides database methods into a single location without the use
 * of globals.
 */
abstract class DatabaseTool implements DatabaseToolInterface {

  /**
   * Schema name validation regular expression.
   */
  public const SCHEMA_NAME_REGEXP =
    '(?=.*\D)[0-9a-z$_\\u0001-\\uFFFF]{1,64}';

  /**
   * Table name validation regular expression.
   */
  public const TABLE_NAME_REGEXP =
    '^(?=.*\D)[0-9a-z$_\\u0001-\\uFFFF]{1,64}';

  /**
   * Instance driver that should be set by the constructor.
   *
   * @var \Drupal\dbxschema\Plugin\CrossSchemaInterface
   */
  protected $driver;

  /**
   * A database Connection related to this tool.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * Reserved schema name patterns.
   *
   * @var array
   */
  protected static $reservedSchemaPatterns;

  /**
   * The Drupal schema name.
   *
   * @var string
   */
  protected static $drupalSchema;

  /**
   * {@inheritdoc}
   */
  public static function getDriverImplementation(
    ?string $driver_name = NULL,
    bool $reset_cache = FALSE,
  ) {
    static $drivers = [];
    // Check if we already loaded the driver.
    if (empty($drivers[$driver_name ?? '']) || $reset_cache) {
      // Nope, try to load it.
      $set_default = FALSE;
      if (empty($driver_name)) {
        // Set default driver using current Drupal connection.
        $set_default = TRUE;
        $connection_options = \Drupal::database()->getConnectionOptions();
        $driver_name = $connection_options['driver'];
      }

      // Get cross-schema implementation for this driver.
      $dbxs_pm = \Drupal::service('plugin.manager.dbxschema');
      $dbxs_plugin = $dbxs_pm->getInstance(
        ['driver' => $driver_name]
      );

      // If we are looking for a default driver, retry with the first available.
      if (empty($dbxs_plugin) && $set_default) {
        $driver_name = array_key_first($dbxs_pm->getDefinitions());
        $dbxs_plugin = $dbxs_pm->getInstance(
          ['driver' => $driver_name]
        );
      }

      // Throw an exception if we didn't find a matching driver.
      if (empty($dbxs_plugin)) {
        if (empty($driver_name)) {
          $message = "No default database cross-schema driver implementation found.";
        }
        else {
          $message = "No database cross-schema implementation found for database driver '$driver_name'.";
        }
        throw new DatabaseToolException($message);
      }

      $drivers[$driver_name] = $dbxs_plugin;
      if ($set_default) {
        $drivers[''] = $dbxs_plugin;
      }
    }

    return $drivers[$driver_name];
  }

  /**
   * {@inheritdoc}
   */
  public static function getDatabaseTool(
    $source = NULL,
  ) {
    if (is_object($source) && ($source instanceof Connection)) {
      $connection = $source;
      if ($source instanceof CrossSchemaConnectionInterface) {
        $source = $connection->getDriver()->driver();
      }
      elseif (!empty($connection->getConnectionOptions()['driver'])) {
        $source = $connection->getConnectionOptions()['driver'];
      }
      else {
        throw new DatabaseToolException('Unable to determine which database driver to use.');
      }
    }
    elseif (!empty($source) && !is_string($source)) {
      // Invalid argument.
      throw new DatabaseToolException('Invalid argument passed to DatabaseTool::getDatabaseTool().');
    }

    try {
      $plugin = static::getDriverImplementation($source);
    }
    catch (DatabaseToolException $e) {
      $message = $e->getMessage() . ' Did you forget to enable the corresponding database cross-schema driver?';
      throw new PluginException($e . $message);
    }
    $database_tool_class = $plugin->getClass('DatabaseTool');
    return new $database_tool_class($connection ?? NULL);
  }

  /**
   * {@inheritdoc}
   */
  public static function getConnection(
    string $schema_name = '',
    $database = 'default',
    ?string $driver_name = NULL,
  ) {

    // Check if a driver was specified.
    if (empty($driver_name)) {
      // No driver specified. Make sure we will use the same driver used by the
      // connection.
      // Check a key was provided instead of a connection object.
      if (is_string($database)) {
        $database = trim($database);
        if ('' == $database) {
          $database = 'default';
        }
        // Get the corresponding connection object.
        try {
          $database = Database::getConnection(
            'default',
            $database
          );
        }
        catch (\RuntimeException $e) {
          throw new ConnectionException($e->getMessage());
        }
      }

      if (empty($database)) {
        $database = \Drupal::database();
      }

      if (!is_a($database, Connection::class)
      ) {
        throw new DatabaseToolException(
          "The provided database is not a Drupal database Connection object or valid Drupal database key."
        );
      }

      $connection_options = $database->getConnectionOptions();
      $driver_name = $connection_options['driver'];
    }

    $plugin = static::getDriverImplementation($driver_name);
    $connection_class = $plugin->getClass('Connection');
    return new $connection_class($schema_name, $database);
  }

  /**
   * Creates a new instance with the appropriate driver and connection.
   *
   * @param ?Connection $connection
   *   A reference connection to use for database operations. It must be
   *   compatible with the DatabaseTool driver implementation.
   *   Default: Drupal database connection.
   */
  public function __construct(?Connection $connection = NULL) {
    // Child class must set their $this->driver and call parent constructor.
    $this->connection = $connection ?? \Drupal::database();
  }

  /**
   * {@inheritdoc}
   */
  abstract public function getDrupalSchemaName() :string;

  /**
   * {@inheritdoc}
   */
  public function isInvalidSchemaName(
    string $schema_name,
    bool $ignore_reservation = FALSE,
    bool $reload_config = FALSE,
  ) :string {

    // @todo Maybe add a flag to enable message translation.
    // Reminder: exception messages should not be translated while user
    // interface should be. Here, we may use the messages in both situation.
    $issue = '';
    // Check reservations.
    if (!$ignore_reservation) {
      // Check reserved patterns.
      // Note: other reserved patterns should be added by other extensions when
      // they are installed, through config modifications.
      // See dbxschema_install() for an example.
      static::initSchemaReservation($reload_config);
      if ($reserved = static::isSchemaReserved($schema_name)) {
        $pattern = array_key_first($reserved);
        $description = $reserved[$pattern];
        $issue =
          "'$schema_name' matches the reservation pattern '$pattern' used for: $description.";
      }
    }
    return $issue;
  }

  /**
   * {@inheritdoc}
   */
  protected function initSchemaReservation(bool $reload_config = FALSE) :void {
    if ($reload_config || !isset(static::$reservedSchemaPatterns)) {
      $reserved_schema_patterns = \Drupal::config('dbxschema.settings')
        ->get('reserved_schema_patterns')
        ?? [];
      static::$reservedSchemaPatterns = $reserved_schema_patterns;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function reserveSchemaPattern(
    string $pat_regex,
    string $description = '',
  ) :void {
    static::initSchemaReservation();
    if (empty($pat_regex)
        || (!preg_match('/[a-z_\\xA0-\\xFF0-9]/i', $pat_regex))
    ) {
      throw new DatabaseToolException("Invalid schema name pattern.");
    }
    static::$reservedSchemaPatterns[$pat_regex] = $description;
  }

  /**
   * {@inheritdoc}
   */
  public function getReservedSchemaPattern() :array {
    static::initSchemaReservation();
    return static::$reservedSchemaPatterns;
  }

  /**
   * {@inheritdoc}
   */
  public function freeSchemaPattern(
    string $pat_regex,
    bool $free_all_matching = FALSE,
  ) :array {
    static::initSchemaReservation();
    $removed_patterns = [];
    if (array_key_exists($pat_regex, static::$reservedSchemaPatterns)) {
      $removed_patterns[$pat_regex] =
        static::$reservedSchemaPatterns[$pat_regex];
      unset(static::$reservedSchemaPatterns[$pat_regex]);
    }
    if ($free_all_matching) {
      foreach (static::$reservedSchemaPatterns as $regex => $reason) {
        $regex_fix = preg_replace('/(?<!\.)\*/', '.*', $regex);
        if (preg_match("/^$regex_fix\$/", $pat_regex)) {
          $removed_patterns[$regex] =
            static::$reservedSchemaPatterns[$regex];
          unset(static::$reservedSchemaPatterns[$regex]);
        }
      }
    }
    return $removed_patterns;
  }

  /**
   * {@inheritdoc}
   */
  public function isSchemaReserved(string $schema_name) {
    static::initSchemaReservation();
    $reserved = FALSE;
    foreach (static::$reservedSchemaPatterns as $reserved_pattern => $description) {
      // Adds regex wildcard.
      $reserved_pattern = preg_replace('/(?<!\.)\*/', '.*', $reserved_pattern);
      if (preg_match("/^$reserved_pattern\$/", $schema_name)) {
        $reserved[$reserved_pattern] = $description;
      }
    }
    return $reserved;
  }

  /**
   * {@inheritdoc}
   */
  abstract public function schemaExists(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :bool;

  /**
   * {@inheritdoc}
   */
  abstract public function createSchema(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * {@inheritdoc}
   */
  abstract public function cloneSchema(
    string $source_schema,
    string $target_schema,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * {@inheritdoc}
   */
  abstract public function renameSchema(
    string $old_schema_name,
    string $new_schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * {@inheritdoc}
   */
  abstract public function dropSchema(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * {@inheritdoc}
   */
  abstract public function getSchemaSize(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :int;

  /**
   * {@inheritdoc}
   */
  abstract public function getDatabaseSize(
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :int;

  /**
   * {@inheritdoc}
   */
  public function parseTableDdl(string $table_ddl) :array {
    $table_definition = [
      'columns' => [],
      'constraints' => [],
      'indexes' => [],
      'dependencies' => [],
    ];
    // Note: if we want to process more exotic table creation strings not
    // comming from ::getTableDdl(), we will have to reformat the
    // string first here.
    $table_raw_definition = explode("\n", $table_ddl);

    // Skip "CREATE TABLE" line.
    $i = 1;
    // Loop until end of table definition.
    while (($i < count($table_raw_definition))
        && (!preg_match('/^\s*\)\s*;\s*$/', $table_raw_definition[$i]))
    ) {
      if (empty($table_raw_definition[$i])) {
        ++$i;
        continue;
      }
      if (
          preg_match(
            '/^\s*CONSTRAINT\s*([\w\$\x80-\xFF\.]+)\s+(.+?),?\s*$/',
            $table_raw_definition[$i],
            $match
          )
      ) {
        // Constraint.
        $constraint_name = $match[1];
        $constraint_def = $match[2];
        $table_definition['constraints'][$constraint_name] = $constraint_def;
        if (preg_match(
              '/
                # Match "FOREIGN KEY ("
                FOREIGN\s+KEY\s*\(
                   # Capture current table columns (one or more).
                  (
                    (?:[\w\$\x80-\xFF\.]+\s*,?\s*)+
                  )
                \)\s*
                # Match "REFERENCES"
                REFERENCES\s*
                  # Caputre evental schema name.
                  ([\w\$\x80-\xFF]+\.|)
                  # Caputre foreign table name.
                  ([\w\$\x80-\xFF]+)\s*
                  \(
                    # Capture foreign table columns (one or more).
                    (
                      (?:[\w\$\x80-\xFF]+\s*,?\s*)+
                    )
                  \)
              /ix',
              $constraint_def,
              $match
            )
        ) {
          $table_columns = preg_split('/\s*,\s*/', $match[1]);
          $foreign_table_schema = $match[2];
          $foreign_table = $match[3];
          $foreign_table_columns = preg_split('/\s*,\s*/', $match[4]);
          if (count($table_columns) != count($foreign_table_columns)) {
            throw new DatabaseToolException("Failed to parse foreign key definition:\n'$constraint_def'");
          }
          else {
            for ($j = 0; $j < count($table_columns); ++$j) {
              $tcol = $table_columns[$j];
              $ftcol = $foreign_table_columns[$j];
              $table_definition['dependencies'][$foreign_table] =
                $table_definition['dependencies'][$foreign_table]
                ?? [];
              $table_definition['dependencies'][$foreign_table][$tcol] = $ftcol;
            }
          }
        }
      }
      elseif (
        preg_match(
          '/^\s*(\w+)\s+(\w+.*?)(\s+NOT\s+NULL|\s+NULL|)(\s+DEFAULT\s+.+?|),?\s*$/',
          $table_raw_definition[$i],
          $match
        )
      ) {
        // Column.
        $table_definition['columns'][$match[1]] = [
          'type'     => $match[2],
          'not null' => (FALSE !== stripos($match[3], 'NOT')),
          'default'  => ($match[4] === '')
            ? NULL
            : preg_replace('/(?:^\s+DEFAULT\s+)|(?:\s+$)/', '', $match[4]),
        ];
      }
      else {
        // If it happens, it means the dbxschema_get_table_ddl() SQL function
        // changed and this script should be adapted.
        throw new DatabaseToolException(
          'Failed to parse unexpected table definition line format for "'
          . $table_raw_definition[0]
          . '": "'
          . $table_raw_definition[$i]
          . '"'
        );
      }
      ++$i;
    }

    // Parses the rest (indexes and comment).
    if (++$i < count($table_raw_definition)) {
      while ($i < count($table_raw_definition)) {
        if (empty($table_raw_definition[$i])) {
          ++$i;
          continue;
        }
        // Parse index name for later comparison.
        if (preg_match(
              '/
                ^\s*
                CREATE\s+
                (?:UNIQUE\s+)?INDEX\s+(?:CONCURRENTLY\s+)?
                (?:IF\s+NOT\s+EXISTS\s+)?
                # Capture index name.
                ([\w\$\x80-\xFF\.]+)\s+
                # Capture table column.
                ON\s+([\w\$\x80-\xFF\."]+)\s+
                # Capture index structure.
                USING\s+(.+);\s*
                $
              /ix',
              $table_raw_definition[$i],
              $match
            )
        ) {
          // Constraint.
          $table_definition['indexes'][$match[1]] = [
            'query' => trim($match[0]),
            'name'  => $match[1],
            'table'  => $match[2],
            'using' => $match[3],
          ];
        }
        elseif (
          preg_match(
            '/^\s*COMMENT\s+ON\s+TABLE\s+\S+\s+IS\s+\'((?:[^\'\\\\]|\\\\.|\'\')*)(\'\s*;\s*|)$/i',
            $table_raw_definition[$i],
            $match
          )
        ) {
          $table_definition['comment'] = $match[1];
          // Complete the comment if needed (multiline comments).
          while (empty($match[2]) && ($i + 1 < count($table_raw_definition))) {
            ++$i;
            preg_match(
              '/^((?:[^\'\\\\]|\\\\.)*)(\'\s*;\s*|)/i',
              $table_raw_definition[$i],
              $match
            );
            $table_definition['comment'] .= "\n" . $match[1];
          }
        }
        else {
          // If it happens, it means the dbxschema_get_table_ddl() SQL function
          // changed and this script should be adapted.
          throw new DatabaseToolException(
            'Failed to parse unexpected table DDL line format for "'
            . $table_raw_definition[0]
            . '": "'
            . $table_raw_definition[$i]
            . '"'
          );
        }
        ++$i;
      }
    }
    return $table_definition;
  }

  /**
   * {@inheritdoc}
   */
  public function parseTableDdlToDrupal(string $table_ddl) :array {
    $table_structure = $this->parseTableDdl($table_ddl);
    // Start with the name of the table.
    $table_def = [];

    // Description.
    if (!empty($table_structure['comment'])) {
      $table_def['description'] = $table_structure['comment'];
    }

    // Columns/fields:
    $table_def['fields'] = [];
    foreach ($table_structure['columns'] as $column => $column_def) {
      $column_def['type'] = trim($column_def['type']);
      $size = NULL;
      $is_int = FALSE;
      $is_float = FALSE;
      // Check for serial.
      if (isset($column_def['default'])
          && preg_match('/^nextval\(/i', $column_def['default'])
      ) {
        unset($column_def['default']);
        if ($column_def['type'] == 'bigint') {
          $column_def['type'] = 'bigserial';
          $size = 'big';
        }
        elseif ($column_def['type'] == 'smallint') {
          $column_def['type'] = 'smallserial';
          $size = 'small';
        }
        else {
          $column_def['type'] = 'serial';
        }
      }
      // Check specified string length or precision and extract it.
      $length = $precision = $scale = NULL;
      if (preg_match('/\(\s*(\d+)\s*\)$/i', $column_def['type'], $match)) {
        $length = intval($match[1]);
        $column_def['type'] = substr(
          $column_def['type'],
          0,
          strpos($column_def['type'], '(')
        );
      }
      elseif (
        preg_match(
          '/\(\s*(\d+)\s*,\s*(\d+)\s*\)$/i',
          $column_def['type'],
          $match
        )
      ) {
        $length = intval($match[1]);
        $scale = intval($match[2]);
        $column_def['type'] = substr(
          $column_def['type'],
          0,
          strpos($column_def['type'], '(')
        );
      }
      // Remove extra stuff from type name.
      $short_type = $column_def['type'];
      $i = strpos($short_type, ' ');
      if (FALSE !== $i) {
        $short_type = substr($short_type, 0, $i);
      }

      // Remap types if needed.
      // Supported types are :
      // 'char', 'varchar', 'text', 'blob', 'int', 'float', 'numeric', 'serial'.
      // @todo add MySQL types.
      $pg_type = NULL;
      switch ($short_type) {
        case 'bigint':
          $column_type = 'int';
          $pg_type = $column_def['type'];
          $size = 'big';
          $is_int = TRUE;
          break;

        case 'int':
        case 'integer':
          $column_type = 'int';
          $pg_type = $column_def['type'];
          $size = 'medium';
          $is_int = TRUE;
          break;

        case 'smallint':
          $column_type = 'int';
          $pg_type = $column_def['type'];
          $size = 'small';
          $is_int = TRUE;
          break;

        case 'boolean':
          $column_type = 'text';
          $pg_type = $column_def['type'];
          break;

        case 'decimal':
          $column_type = 'float';
          $pg_type = $column_def['type'];
          $precision = $length;
          $length = NULL;
          $is_float = TRUE;
          break;

        case 'real':
          $column_type = 'float';
          $pg_type = $column_def['type'];
          $is_float = TRUE;
          break;

        case 'double':
          $column_type = 'float';
          $pg_type = $column_def['type'];
          $size = 'big';
          $is_float = TRUE;
          break;

        case 'bigserial':
          $column_type = 'serial';
          $pg_type = $column_def['type'];
          $size = 'big';
          $is_int = TRUE;
          break;

        case 'serial':
          $column_type = 'serial';
          $pg_type = $column_def['type'];
          $size = 'medium';
          $is_int = TRUE;
          break;

        case 'smallserial':
          $column_type = 'serial';
          $pg_type = $column_def['type'];
          $size = 'small';
          $is_int = TRUE;
          break;

        case 'character':
          if (FALSE !== stripos($column_def['type'], 'var')) {
            $column_type = 'varchar';
          }
          else {
            $column_type = 'char';
          }
          $pg_type = $column_def['type'];
          break;

        case 'bytea':
          $column_type = 'blob';
          $pg_type = $column_def['type'];
          break;

        default:
          // Date/time, money, enumerated, geometric, network address, etc.
          $column_type = 'text';
          $pg_type = $column_def['type'];
          break;
      }

      $table_def['fields'][$column] = [
        'type' => $column_type,
        'not null' => $column_def['not null'],
      ];
      if (!empty($pg_type)) {
        $table_def['fields'][$column]['pgsql_type'] = $pg_type;
      }
      if (isset($length)) {
        $table_def['fields'][$column]['length'] = $length;
      }
      if (isset($size)) {
        $table_def['fields'][$column]['size'] = $size;
      }
      if (isset($precision)) {
        $table_def['fields'][$column]['precision'] = $precision;
      }
      if (isset($scale)) {
        $table_def['fields'][$column]['scale'] = $scale;
      }
      if (isset($column_def['default'])) {
        if ($is_int) {
          $table_def['fields'][$column]['default'] = (int) $column_def['default'];
        }
        elseif ($is_float) {
          $table_def['fields'][$column]['default'] = (float) $column_def['default'];
        }
        else {
          $table_def['fields'][$column]['default'] = $column_def['default'];
        }
      }
    }

    // Constraints.
    foreach ($table_structure['constraints'] as $constraint => $cdef) {
      $cdef = trim($cdef);
      if (preg_match('/^PRIMARY\s+KEY\s+\((.+)\)/i', $cdef, $match)) {
        $table_def['primary key'] = preg_split('/\s*,\s*/', $match[1]);
      }
      elseif (preg_match('/^UNIQUE\s+\((.+)\)/i', $cdef, $match)) {
        if (!array_key_exists('unique keys', $table_def)) {
          $table_def['unique keys'] = [];
        }
        $table_def['unique keys'][$constraint] =
          preg_split('/\s*,\s*/', $match[1]);
      }
      elseif (preg_match(
        '/
          # Match "FOREIGN KEY ("
          FOREIGN\s+KEY\s*\(
             # Capture current table columns (one or more).
            (
              (?:[\w\$\x80-\xFF\.]+\s*,?\s*)+
            )
          \)\s*
          # Match "REFERENCES"
          REFERENCES\s*
            # Caputre evental schema name.
            ([\w\$\x80-\xFF]+\.|)
            # Caputre foreign table name.
            ([\w\$\x80-\xFF]+)\s*
            \(
              # Capture foreign table columns (one or more).
              (
                (?:[\w\$\x80-\xFF]+\s*,?\s*)+
              )
            \)
        /ix',
        $cdef,
        $match
      )) {
        if (!array_key_exists('foreign keys', $table_def)) {
          $table_def['foreign keys'] = [];
        }
        $table_columns = preg_split('/\s*,\s*/', $match[1]);
        $foreign_table_schema = $match[2];
        $foreign_table = $match[3];
        $foreign_table_columns = preg_split('/\s*,\s*/', $match[4]);
        $table_def['foreign keys'][$constraint] = [
          'table' => $foreign_table,
          'columns' => [],
        ];
        if (count($table_columns) == count($foreign_table_columns)) {
          for ($col = 0; $col < count($table_columns); ++$col) {
            $tcol = $table_columns[$col];
            $ftcol = $foreign_table_columns[$col];
            $table_def['foreign keys'][$constraint]['columns'][$tcol] = $ftcol;
          }
        }
      }
    }
    return $table_def;
  }

}
