<?php

namespace Drupal\dbxschema\Database;

/**
 * This interface defines methods of DatabaseTool class.
 */
interface DatabaseToolInterface {

  /**
   * Returns the cross-schema plugin for the given driver.
   *
   * @param string $driver_name
   *   Driver name (ex.: 'pgsql' or 'mysql').
   *   If no driver is specified, it will use current Drupal database driver
   *   name and if there is no corresponding driver, it will try to use the
   *   first available one.
   *   Default: NULL.
   * @param bool $reset_cache
   *   If TRUE, local driver cache is ignored and refreshed.
   *
   * @return \Drupal\dbxschema\Plugin\CrossSchemaInterface
   *   A plugin implementation for the given driver.
   *
   * @throws \Drupal\dbxschema\Exception\DatabaseToolException
   *   If no implementation is available for the given driver.
   */
  public static function getDriverImplementation(
    ?string $driver_name = NULL,
    bool $reset_cache = FALSE,
  );

  /**
   * Returns an implementation of DatabaseTool class for the given driver.
   *
   * This static method selects the appropriate driver corresponding to the
   * given source. If the source is a Connection object, that object will be
   * provided to the driver DatabaseTool constructor.
   *
   * @param \Drupal\dbxschema\Database\ConnectionInterface|string|null $source
   *   Can be either a driver name (ex.: 'pgsql' or 'mysql') or a
   *   \Drupal\dbxschema\Database\ConnectionInterface instance. The source will
   *   be used to provide the matching DatabaseTool.
   *   If no driver is specified, it will use current Drupal database driver
   *   name and if there is no corresponding driver, it will try to use the
   *   first available one.
   *   Default: NULL.
   *
   * @return \Drupal\dbxschema\Database\DatabaseTool
   *   A database tool implementation for the given driver.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *   If no implementation is available for the given driver.
   * @throws \Drupal\dbxschema\Exception\DatabaseToolException
   *   If the given parameter is not null, not a string and not an instance of
   *   \Drupal\dbxschema\Database\ConnectionInterface.
   */
  public static function getDatabaseTool(
    ConnectionInterface|string|null $source = NULL,
  );

  /**
   * Returns an implementation of CrossSchemaConnection class.
   *
   * @param string $schema_name
   *   The schema name to use.
   *   Default: '' (no schema). It will throw exceptions on methods needing a
   *   default schema but may work on others or when a schema can be passed
   *   as parameter.
   * @param \Drupal\Core\Database\Connection|string $database
   *   Either a \Drupal\Core\Database\Connection instance or a
   *   Drupal database key string (from current site's settings.php).
   *   Extra databases specified in settings.php do not need to specify a
   *   schema name as a database prefix parameter. The prefix will be managed by
   *   this connection class instance.
   * @param string $driver_name
   *   Driver name (ex.: 'pgsql' or 'mysql').
   *   If no driver is specified, it will use current Drupal database driver
   *   name.
   *   Default: NULL.
   *
   * @return \Drupal\dbxschema\Database\CrossSchemaConnection
   *   A cross-schema connection implementation for the given driver.
   *
   * @throws \Drupal\dbxschema\Exception\DatabaseToolException
   *   If no implementation is available for the given driver.
   * @throws \Drupal\dbxschema\Exception\ConnectionException
   *   If the database connection fails.
   */
  public static function getConnection(
    string $schema_name = '',
    $database = 'default',
    ?string $driver_name = NULL,
  );

  /**
   * Get Drupal schema name.
   *
   * Use:
   * @code
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $drupal_schema = $db_tool->getDrupalSchemaName();
   * @endcode
   *
   * @return string
   *   The name of the schema used by Drupal installation or an empty string if
   *   Drupal is stored in a different database.
   */
  public function getDrupalSchemaName() :string;

  /**
   * Check that the given schema name is a valid schema name.
   *
   * Schema name validation can be altered through the configuration variable
   * reserved_schema_patterns of dbxschema.settings. This configuration
   * variable contains a list of regex with thier description, used to reserve
   * schema name patterns. For instance, the key '_chado*' with the value
   * 'external (non-Drupal) Chado instances' will make this function returns a
   * issue message saying that the pattern is reserved for 'external
   * (non-Drupal) chado instances' when a schema name '_chado_beta' is checked.
   *
   * Extending modules willing to reserve schema names should use something
   * similar to the following code in their "<module name>.install" file:
   * @code
   *   function <module name>_install($is_syncing) {
   *     // Reserves 'myschema' schema in 'reserved_schema_patterns' settings.
   *     $config = \Drupal::service('config.factory')
   *       ->getEditable('dbxschema.settings')
   *     ;
   *     $reserved_schema_patterns = $config->get('reserved_schema_patterns') ?? [];
   *     $reserved_schema_patterns['myschema'] = 'my schema';
   *     $config->set('reserved_schema_patterns', $reserved_schema_patterns)->save();
   *   }
   *
   *   function <module name>_uninstall() {
   *     // Unreserves 'myschema' schemas in 'reserved_schema_patterns' settings.
   *     $config = \Drupal::service('config.factory')
   *       ->getEditable('dbxschema.settings')
   *     ;
   *     $reserved_schema_patterns = $config->get('reserved_schema_patterns') ?? [];
   *     unset($reserved_schema_patterns['myschema']);
   *     $config->set('reserved_schema_patterns', $reserved_schema_patterns)->save();
   *   }
   * @endcode
   *
   * Use:
   * @code
   *   $schema_name = 'name_to_check';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   if ($issue = $db_tool->isInvalidSchemaName($schema_name)) {
   *     throw new \Exception('Invalid schema name: ' . $issue);
   *   }
   * @endcode
   *
   * @param string $schema_name
   *   The name of the schema to validate.
   * @param bool $ignore_reservation
   *   If TRUE, reserved schema names are considered as valid.
   *   Default: FALSE.
   * @param bool $reload_config
   *   Forces schema reserved names config reloading.
   *   Default: FALSE.
   *
   * @return string
   *   An empty string if the schema name is valid or a string describing the
   *   issue in the name otherwise.
   */
  public function isInvalidSchemaName(
    string $schema_name,
    bool $ignore_reservation = FALSE,
    bool $reload_config = FALSE,
  ) :string;

  /**
   * Adds a schema name pattern for reservation.
   *
   * Schema names matching the given pattern will be considered invalid by
   * ::isInvalidSchemaName and will not be allowed in Connection or Schema
   * objects.
   *
   * @param string $pat_regex
   *   A simple schema name or a regular expression. Do not include regex
   *   delimiters nor starting '^' and ending '$' in the expression as they will
   *   be automatically added by the check system. Note that the '*' sign not
   *   preceded by a dot will be replaced by '.*'. It simplifies the way
   *   schema patterns can be defined by non-regex aware persons. If you need
   *   to use the '*' quantifier for a specific character, replace it by '{0,}'.
   *   ex.: 'internal_schemax{0,}' would match 'internal_schema' and
   *   'internal_schemaxxx' while 'internal_schemax*' would also match
   *   'internal_schemaxabcd'.
   * @param string $description
   *   The description of the reservation that may be displayed to users when a
   *   schema name is denied.
   *
   * @throws \Drupal\dbxschema\Exception\DatabaseToolException
   *   If the pattern is empty or does not contain any valid schema name
   *   character.
   */
  public function reserveSchemaPattern(
    string $pat_regex,
    string $description = '',
  ) :void;

  /**
   * Returns the list of reserved schema name pattern.
   *
   * @return array
   *   The list of reserved patterns as keys and their associated descriptions
   *   as values.
   */
  public function getReservedSchemaPattern() :array;

  /**
   * Removes a schema name reservation pattern from the list.
   *
   * @param string $pat_regex
   *   The regular expression to remove from the list if it is there.
   * @param bool $free_all_matching
   *   If TRUE, the provided pattern will be considered as a regular string with
   *   no special characters meaning and any current pattern matching that
   *   string will be removed form current reservation list.
   *
   * @return array
   *   Returns an associative array containing the removed patterns as keys and
   *   their associated descriptions as values. An empty array if no pattern has
   *   been removed.
   */
  public function freeSchemaPattern(
    string $pat_regex,
    bool $free_all_matching = FALSE,
  ) :array;

  /**
   * Tells if a schema name is reserved or not.
   *
   * @param string $schema_name
   *   The name of the schema to check.
   *
   * @return bool|array
   *   FALSE if the given schema name is not reserved, otherwise it will return
   *   an array with the reservation pattern matching the name as keys and their
   *   associated descriptions as values.
   */
  public function isSchemaReserved(string $schema_name);

  /**
   * Check that the given schema exists.
   *
   * Use:
   * @code
   *   $schema_name = 'name_to_test';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   if ($db_tool->schemaExists($schema_name)) {
   *     // Schema exists.
   *   }
   * @endcode
   *
   * @param string $schema_name
   *   Schema name.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @return bool
   *   TRUE/FALSE depending upon whether or not the schema exists.
   */
  public function schemaExists(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :bool;

  /**
   * Creates the given schema.
   *
   * The schema to create must not exist. If an error occurs, an exception
   * is thrown.
   *
   * Use:
   * @code
   *   $schema_name = 'name_to_create';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $db_tool->createSchema($schema_name);
   * @endcode
   *
   * @param string $schema_name
   *   Name of schema to create.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @throws \Drupal\Core\Database\DatabaseExceptionWrapper
   */
  public function createSchema(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * Clones a schema into new (unexisting) one.
   *
   * The target schema must not exist.
   *
   * @code
   *   $source_schema_name = 'source';
   *   $target_schema_name = 'target';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $db_tool->cloneSchema($source_schema_name, $target_schema_name);
   * @endcode
   *
   * @param string $source_schema
   *   Source schema to clone.
   * @param string $target_schema
   *   Destination schema that will be created and filled with a copy of
   *   $source_schema.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A cross database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @throws \Drupal\Core\Database\DatabaseExceptionWrapper
   */
  public function cloneSchema(
    string $source_schema,
    string $target_schema,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * Renames a schema.
   *
   * The new schema name must not be used by an existing schema. If an error
   * occurs, an exception is thrown.
   *
   * @code
   *   $old_schema_name = 'old';
   *   $new_schema_name = 'new';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $db_tool->renameSchema($old_schema_name, $new_schema_name);
   * @endcode
   *
   * @param string $old_schema_name
   *   The old schema name to rename.
   * @param string $new_schema_name
   *   New name to use.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A cross database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @throws \Drupal\Core\Database\DatabaseExceptionWrapper
   */
  public function renameSchema(
    string $old_schema_name,
    string $new_schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * Removes the given schema.
   *
   * The schema to remove must exist. If an error occurs, an exception is
   * thrown.
   *
   * @code
   *   $schema_name = 'schema_to_delete';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $db_tool->dropSchema($schema_name);
   * @endcode
   *
   * @param string $schema_name
   *   Name of schema to remove.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A cross database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @throws \Drupal\Core\Database\DatabaseExceptionWrapper
   */
  public function dropSchema(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :void;

  /**
   * Returns the size in bytes of a PostgreSQL schema.
   *
   * @code
   *   $schema_name = 'schema';
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $schema_size = $db_tool->getSchemaSize($schema_name);
   * @endcode
   *
   * @param string $schema_name
   *   Schema name.
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A cross database connection object.
   *   If NULL, current Drupal database is used.
   *
   * @return int
   *   The size in bytes of the schema or 0 if the size is not available.
   *
   * @throws \Drupal\dbxschema\Exception\DatabaseToolException
   */
  public function getSchemaSize(
    string $schema_name,
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :int;

  /**
   * Returns the size in bytes of a database.
   *
   * @code
   *   $db_tool = \Drupal::service('dbxschema.tool');
   *   $db_size = $db_tool->getDatabaseSize();
   * @endcode
   *
   * @param \Drupal\dbxschema\Database\CrossSchemaConnectionInterface $db
   *   A cross database connection object.
   *
   * @return int
   *   The size in bytes of the database or 0 if the size is not available.
   */
  public function getDatabaseSize(
    ?CrossSchemaConnectionInterface $db = NULL,
  ) :int;

  /**
   * Turns a table DDL string into a more usable structure.
   *
   * @param string $table_ddl
   *   A string containing table definition as returned by
   *   \Drupal\dbxschema\Database\CrossSchemaSchemaInterface::getTableDdl().
   *
   * @returns array
   *   An associative array with the following structure:
   *   @code
   *   [
   *     'columns' => [
   *       <column name> => [
   *        'type'     => <column type>,
   *        'not null' => <TRUE if column cannot be NULL, FALSE otherwise>,
   *        'default'  => <column default value or NULL for no default>,
   *       ],
   *       ...
   *     ],
   *     'constraints' => [
   *       <constraint name> => <constraint definition>,
   *       ...
   *     ],
   *     'indexes' => [
   *       <index name> => [
   *         'query' => <index creation query>,
   *         'name'  => <index name>,
   *         'table' => <'table.column' names owning the index>,
   *         'using' => <index type/structure>,
   *       ],
   *       ...
   *     ],
   *     'comment' => <table description>,
   *     'dependencies' => [
   *       <foreign table name> => [
   *         <this table column name> => <foreign table column name>,
   *         ...
   *       ],
   *       ...
   *     ],
   *   ];
   *   @endcode
   */
  public function parseTableDdl(string $table_ddl) :array;

  /**
   * Parses a table DDL and returns a Drupal schema definition.
   *
   * An exception is thrown if the table is not found.
   *
   * @param string $table_ddl
   *   A string containing table definition as returned by
   *   \Drupal\dbxschema\Database\CrossSchemaSchemaInterface::getTableDdl().
   *
   * @return array
   *   An array with details of the table reflecting what is in database.
   */
  public function parseTableDdlToDrupal(string $table_ddl) :array;

}
